#! /usr/bin/perl

# GWSD.pl
#
# Graph based Word Sense Disambiguation
# 
# Author: Ravi S Sinha
# University of North Texas
# Written: Fall 2006
# Updated: Summer 2007

use Getopt::Long; # CPAN module, for long command-line options
use Storable; # CPAN module, for storing hashes for quick retrieval
use lib './';
use myHeader; # Own module, contains all include files
use strict; # Pragma, enforces strict mode
use warnings; # Pragma, enforces warnings mode
# LOG file opened in the header file in append mode instead so that all
# modules can write onto it

# variables
my (@discourse, @wordObj, @sentenceObj); # The objects
my ($corpus, $windowSize, $partOfSpeech, $p);
my ($choice, $similarityMeasure, $pathToCorpus);
my ($metric, $folderName, $featureFile, $verbose);
my ($indegreeFeatureFile, $TextRankFeatureFile, $BetweennessFeatureFile, $ClosenessFeatureFile);
my $dataDIR;
my $senseGraph;
my @sentLOCAL;

unless (scalar(@ARGV)) {
	print "\nGWSD: Graph-based Word Sense Disambiguation

Usage: Choose one of the following options

-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

STEP 1: If you want to build the graphs for a new data set

\$ perl GWSD.pl -choice 1 -pos [n | v | a | r | ALL] \ 
                -sim [lesk | lch | wup | lin | jcn | res | ALL] \
	        -window window_size -corpus [Senseval-2 | Senseval-3 | corpus_name] \
                -path pathToFolderContainingCorpus (only if corpus is 
                      neither Senseval-2 nor Senseval-3)\
                -v (only if you want it to be verbose) 

Whenever you want to build a new set of graphs with the exact same specifications, 
make sure to delete the old folder of graph files.

OR  

-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

STEP 2: Use the output from STEP 1 to run the disambiguation algorithm

\$ perl GWSD.pl -choice 2 -pos [n | v | a | r | ALL] \
                -sim [lesk | lch | wup | lin | jcn | res | ALL] \
                -window window_size -corpus [Senseval-2 | Senseval-3 | corpus_name] \
                -metric [indegree | TextRank | Betweenness | Closeness] \
                -featureFile name_of_file \
                -path pathToFolderContainingCorpus (only if corpus is 
                      neither Senseval-2 nor Senseval-3)

OR 

-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=-=

STEP 3: Use the output from STEP 2 to run a voting across feature files

\$ perl GWSD.pl -choice 3 -corpus [Senseval-2 | Senseval-3 | corpus_name] \
                -path pathToFolderContainingCorpus (only if the corpus is 
                      neither Senseval-2 nor Senseval-3) \
                -inf indegreeFeatureFile \
                -trf TextRankFeatureFile \
                -clf ClosenessFeatureFile (optional) \
                -bwf BetweennessFeatureFile (optional)
	\n\n";	

	exit(1);
}


GetOptions (	'corpus=s', \$corpus, 
		'path=s', \$pathToCorpus,
		'window=i', \$windowSize,
		'pos=s', \$partOfSpeech,
		'sim=s', \$similarityMeasure,
		'metric=s', \$metric,
		'choice=i', \$choice,
		'folder=s', \$folderName,
		'featureFile=s', \$featureFile,
		'trf=s', \$TextRankFeatureFile,
		'inf=s', \$indegreeFeatureFile,
		'bwf=s', \$BetweennessFeatureFile,
		'clf=s', \$ClosenessFeatureFile,		
		'v', \$verbose
);


system('clear');

# Adding a trailing / in the paths for consistency
if(defined($pathToCorpus)) {
  if($pathToCorpus =~ /\/$/) {} # We are okay
  else {
    $pathToCorpus = $pathToCorpus."/"; # Add a trailing '/'
  }
}

if(!defined($choice)) { print "\nError: please specify one of the three available choices ... \n"; exit(1); }

if($choice<1 || $choice>3) { print "Invalid option for choice, please try again ... \n"; exit(1); }
print "Thanks for using GWSD. This software is the property of the University of North Texas.\n"; 
if($choice==1) {

print "\nYou have chosen option number 1.\n";

if(($similarityMeasure ne 'lesk') && ($similarityMeasure ne 'jcn') && ($similarityMeasure ne 'res') && ($similarityMeasure ne 'lin') && ($similarityMeasure ne 'lch') && ($similarityMeasure ne 'wup') && ($similarityMeasure ne 'ALL')) {
	print "Invalid option $similarityMeasure ... \n"; exit(1);
}


if(!defined($partOfSpeech) || !defined($similarityMeasure) || !defined($windowSize) || !defined($corpus)) {
	print "\nSorry, your input is missing some important components. Please refer to the README for more information ... \n";
	exit(1);
}

# Time to open data files
# For Senseval-2 and -3 the path is stored in the file GWSD.conf
# This file needs to be changed
# For other datasets, a path to the folder containing the dataset must be input by the user

if(($corpus eq 'Senseval-2') || ($corpus eq 'Senseval-3')) {
$pathToCorpus = "./";
open DATAIN, "GWSD.conf";
while(my $line = <DATAIN>) {
	my @arr = split /\s+/, $line;
	if($corpus eq $arr[0]) {
		$dataDIR = $arr[1];
	}
}
close DATAIN;
}
elsif ($pathToCorpus ne './') {
	$dataDIR = $pathToCorpus;
}
else { 
	$dataDIR = "don't go anywhere";
	print "\nSorry, the dataset is apparently not located in a folder.\n";
	exit(1);
}
# Get all files inside the folder one by one
opendir(DIR, $dataDIR) or die "Could not open dir -> the folder $pathToCorpus is either empty or not accessible\nSystem says: $!";
my @docs=grep($_ ne '.' && $_ ne '..' && -f "$dataDIR\/$_", readdir(DIR));

if(($windowSize<0) || ($windowSize>12)) {
	print "Window size too small (invalid) or too large (will take forever to execute) ... \n"; exit(1);
}

if(($partOfSpeech ne 'n') && ($partOfSpeech ne 'v') && ($partOfSpeech ne 'a') && ($partOfSpeech ne 'r') && ($partOfSpeech ne 'ALL')) {
	print "Invalid option $partOfSpeech... \n"; exit(1);
}
$p = $partOfSpeech;

print "All required modules and libraries included. . .\n";

# Call the module ObjectBuilder to 
#	- read all data from the test corpus
#	- create objects of all three classes 
# No explicit return; references to the arrays are being passed here

print "Calling module ObjectBuilder to read all data and to create objects \nbelonging to the three classes. . .\n";
print "Building a graph of senses. . .\n";
print "An object of module Graph created; \nCalling module GraphBuilder to populate the graph. . .\n";

# Creating the folder inside which all the graphs will be stored
system ("mkdir $corpus.$similarityMeasure.$partOfSpeech.$windowSize.Graphs");
my $graphFolder = "$corpus.$similarityMeasure.$partOfSpeech.$windowSize.Graphs";

# Checkpoint 1 - see if everything has been assiged proper values
print LOG "Checkpoint 1\n";
print LOG "GWSD.pl, inside choice 1\n";
print LOG "-" x scrWid. "\n";
print LOG "Corpus = $corpus\tSim measure = $similarityMeasure\tPOS = $partOfSpeech\tWindow = $windowSize\n";
print LOG "Choice = $choice\tPath = $pathToCorpus\tGraphFolder = $graphFolder\n";
print LOG "DataDir = $dataDIR\n";
print LOG "-" x scrWid. "\n"; 

print LOG "Checkpoint 2\n";
print LOG "The files that are being considered in the corpus\n";
print LOG "-" x scrWid ."\n";

# This will generate graphs for all the three data files in the Senseval-2 or Senseval-3
my $numberOfFiles = scalar(@docs);
print LOG "The number of files: $numberOfFiles\n";
my $counter = 0;

foreach my $file(@docs) {
	
	my $f = "$dataDIR".$file;
	# Calling modules from ObjectBuilder.pm
	
	# Creating a local scope does not solve the problem; 
	# Set each object manually to 0 (a routine similar to createObjects)
	
	foreach my $dummy(@sentLOCAL) {$dummy = ''; }
	@sentLOCAL = ();

	foreach my $dumm(@discourse) {$dumm = ''; }
	@discourse = ();
			
	&readinText(\@discourse, $f, \@wordObj);
	&createObjects(\@discourse, \@sentLOCAL); 
	
	print LOG $f. "\t" .$file. "\n";
	print LOG "Name of the array = \@sentLOCAL Length of the array = ".scalar(@sentLOCAL)."\n";

	# Now the arrays are populated with the objects of the classes,
	# so we call the buildGraph subroutine to build the graphs 
	print "\nPlease wait... \n";
	
	$senseGraph = Graph::Undirected->new();
	
	$senseGraph = &buildGraph($graphFolder, $senseGraph, $verbose, $windowSize, $similarityMeasure, $partOfSpeech, \@sentLOCAL);
	push @sentenceObj, @sentLOCAL; 
	
} 

print LOG "\n";
print LOG "-" x scrWid ."\n";
print "Done. Please see the objects in File.Objects and the edges of the graphs\n in the corresponding folder: Corpus.Sim.POS.Window.Graphs\n";

# Storing the objects for future use
store \@sentenceObj, "./Objects/Obj.$corpus.$similarityMeasure.$partOfSpeech.$windowSize";

exit(0);
} # End of choice 1

if($choice==2) {

print "You have chosen option number 2\n";

if(($metric ne 'indegree') && ($metric ne 'TextRank') && ($metric ne 'Closeness') && ($metric ne 'Betweenness')) {
	print "Invalid option $metric ...\n"; exit(0);
}

if(!defined($partOfSpeech) || !defined($similarityMeasure) || !defined($windowSize) || !defined($corpus) || !defined($metric) || !defined($featureFile) ) {
	print "\nSorry, your input is apparently missing some important components. Please refer to the README for a remedy ...\n";
	exit(1);
}

$folderName = "$corpus.$similarityMeasure.$partOfSpeech.$windowSize.Graphs";

if(!(-d $folderName)) { # The so-called X-test for files: will check if the name is a folder
	print "Invalid folder name ... $folderName\nSystem says: $!\n"; exit(0);
}

if(($corpus eq 'Senseval-2') || ($corpus eq 'Senseval-3')) {
open DATAIN, "GWSD.conf";
while(my $line = <DATAIN>) {
	my @arr = split /\s+/, $line;
	if($corpus eq $arr[0]) {
		$dataDIR = $arr[1];
	}
}
close DATAIN;
}
elsif(defined($pathToCorpus)) {
	$dataDIR = $pathToCorpus;
}
else {	
	print "Error: you did not provide the name of the folder where the corpus is located ...\n";
	exit(1);
}
# None of these methods check the accuracy; they just produce the corresponding feature files
# Need to call a generic method to calculate the PRF (from inside these individual methods)

print LOG "Checkpoint 3\nInside option 2\n";
print LOG "-" x scrWid ."\n";
print LOG "Metric = $metric\tFolderName = $folderName\n";
print LOG "FeatureFile = $featureFile\tDataDIR = $dataDIR\n";
print LOG "-" x scrWid ."\n";  

if($metric eq 'indegree') {
	my $sentenceObj = retrieve("./Objects/Obj.$corpus.$similarityMeasure.$partOfSpeech.$windowSize");	
	# returns a reference to the hash table that we created earlier
	&getIndegree($featureFile, $dataDIR, $sentenceObj);  # This is easier because everything is in the objects
}

if($metric eq 'TextRank') {&getTextRank($corpus, $folderName, $featureFile, $dataDIR); }
if($metric eq 'Closeness') {&getCloseness($corpus, $folderName, $featureFile, $dataDIR); }
if($metric eq 'Betweenness') {&getBetweenness($corpus, $folderName, $featureFile, $dataDIR); }
print "\nThe $metric scores have been calculated, and the feature file ./Features/$featureFile is generated.";
print "\nThe precision, recall and F scores are to be found in files ./Scores/$featureFile.Scores.\n";
exit(0);
}

if($choice==3) {

print "You have chosen option number 3\n";

if( !defined($corpus) || !defined($indegreeFeatureFile) || !defined($TextRankFeatureFile) ) {
	print "\nSorry, your input is apparently missing some important components. Please refer to the README for a remedy ...\n";
	exit(1);
}

print LOG "Checkpoint 4\nInside option 3\n";
print LOG "-" x scrWid ."\n";
if(defined($BetweennessFeatureFile)) {
print LOG "Feature files = $indegreeFeatureFile, $TextRankFeatureFile, $BetweennessFeatureFile, $ClosenessFeatureFile\n";
}
else {
print LOG "Feature files = $indegreeFeatureFile, $TextRankFeatureFile\n";
}
print LOG "-" x scrWid ."\n";
if(defined($BetweennessFeatureFile)) {
&vote($indegreeFeatureFile, $TextRankFeatureFile, $BetweennessFeatureFile, $ClosenessFeatureFile, $corpus, $pathToCorpus);
}
else {
&vote2($indegreeFeatureFile, $TextRankFeatureFile, $corpus, $pathToCorpus);
}
print "Voting done. The results are to be found in the file ./Scores/Voting.Scores\n\n";
exit(0);
}
